import reiz
from pytest import fixture
from tempfile import TemporaryDirectory, NamedTemporaryFile
import wave
import random
import struct
from pathlib import Path


def test_beep():
    reiz.audio.Hertz(volume=0, duration_in_ms=100).play()


def test_audio_library():
    for item in reiz.audio.library.__dict__.values():
        item.volume = 0
        item.play_blocking()


@fixture
def mock_wavfiles():
    SAMPLE_LEN = 44100
    with TemporaryDirectory() as fld:
        wf = (Path(fld) / "tmp").with_suffix(".wav")
        with wave.open(str(wf), 'wb') as wfile:
            wfile.setparams((2, 2, 44100, 0, 'NONE', 'not compressed'))
            for i in range(0, SAMPLE_LEN):
                value = random.randint(-32767, 32767)
                packed_value = struct.pack('h', value)
                wfile.writeframes(packed_value)
                wfile.writeframes(packed_value)
            wfile.close()
            yield fld, Path(wf.name).stem
        wf.unlink()


def test_read_folder(mock_wavfiles):
    fld, fname = mock_wavfiles
    lib = reiz.audio.read_folder(fld)
    assert fname in lib.__dict__.keys()



import reiz

# start the MarkerServer which distributes markerstrings over LSL
reiz.marker.start()

# create a window
canvas = reiz.Canvas()

# initialize a clock for timing control
clock = reiz.Clock()
# initialize Cues
# each Cue reveices information about
# the Window where it will be shown -> canvas
# auditory and visual stimuli, e.g. a tone (Hertz) or text (Mural)
# and a markerstring to be send via LSL when .show() is called
hello = reiz.Cue(canvas,
                 audiostim=reiz.audio.Hertz(
                     frequency=400, duration_in_ms=1000),
                 visualstim=reiz.visual.Mural('Hello World!'),
                 markerstr='hello')

# there is also a library of typical auditory and visual stimuli and
# we can for example, take the visual "los" and the auditory "beep" stimuli
# for convenience, we use the text of the go stimulus as marker message
los = reiz.Cue(canvas,
               audiostim=reiz.audio.library.beep,
               visualstim=reiz.visual.library.go,
               markerstr=reiz.visual.library.go.text)

# here we show a fixation cross but don't want to play an auditory stimulus.
# we can therefore either set it to None or just leave it out (defaults to None)
fix = reiz.Cue(canvas,
               visualstim=reiz.visual.library.fixation,
               markerstr='Fixation')

shape = reiz.Cue(canvas,
                 visualstim=reiz.visual.Trapezoid(xpos=(-.25, -.33, .33, .25),
                                                  ypos=(-.25, .25), color='red'),
                 markerstr='Shape')

# we can also use an iterable for the visual stimuli, and they are overlayed
# from left to right
farewell = reiz.audio.Message("Auf Wiedersehen!")
overlay = reiz.Cue(canvas,
                   audiostim=farewell,
                   visualstim=[
                       reiz.visual.library.logo,
                       reiz.visual.Mural('Good Bye', fontsize=1),
                   ],
                   markerstr='Overlayed')

# we create a named variable for a visual stimulus to allow later updating
# because we will repeatedly call ball.show(), we don't use the markerstr
# as these would be otherwise send everytime we .show()
dynamic_ball = reiz.visual.Circle(zoom=1, color='red')
ball = reiz.Cue(canvas, visualstim=dynamic_ball)

# we open a window, show the cues and close the window again
canvas.open()

# we can either add a sleep function afterwards
fix.show()
clock.sleep(1)
# or use the duration parameter. this has the advantage that the window stays
# responsive under Win10 and can e.g. be moved around or resized
hello.show(duration=1)
los.show(duration=1)
shape.show(duration=1)
# we continually update the size of the visualstim of the cue
# because the Cue has a reference to the dynamic_ball, it takes
# over any changes
clock.tick()
# reset the clock to be able to debias sleep towards this time-point
for i in range(0, 100, 1):

    # we update the size of the ball
    dynamic_ball.zoom = i/20
    # and show it
    ball.show()
    # if the show duration is faster than the framerate of your graphics card
    # and screen, expect that the update rate is only as fast as your hardware
    # allows, i.e. with 60 Hz is ~16ms. Sleeping for longer than the minimum
    # framerate ensures controlled calculation time, but frames are still
    # refreshed according to the whims of your graphics card.
    # generally, when you use dynamic cues, e.g. for feedback, consider that
    # there might be a variable amount of time spend for calculation.
    # therefore, it is best practice to sleep controlling for this bias
    # instead of simply sleep a fixed amount of time
    clock.sleep_debiased(0.05)
    # this is roughly equivalent to
    # ```
    # t0 = time.time()
    # execute_task_with_random_runtime()
    # while (time.time()-t0) < 0.05:
    #     pass
    # ```


# we present the cue as long as the audio is playing
overlay.show(duration=farewell.duration)

canvas.close()

reiz.marker.stop()



# -*- coding: utf-8 -*-
"""Standalone MarkerServer

LabRecorder can only start recording Outlets that are existing. If you start a MarkerServer immediatly before you run the experiment, there is not much time for the Recorder to detect the stream.

We there recommend to use a reiz-marker-Server as an independent process. You can either start the MarkerServer as an independent process with `reiz-marker` or `python -m reiz.marker` from your terminal. Shut it down gracefull with `reiz-marker --kill`.

Alternatively, :meth:`~.reiz._marker.safeguard.start` starts such a process from within Python, and you can kill it later with :meth:`~.reiz._marker.safeguard.stop`.

This MarkerServer opens an Outlet that can be detected independently from the experiments you are running. When you then run an experiment, it receives messages from this experiment, and redistributes them in LSL-format.

run from terminal with

.. code-block:: bash

    usage: reiz-marker [-h] [--port PORT] [--host HOST] [--name NAME] [--ping] [--kill]

    Reiz Marker Server

    optional arguments:
    -h, --help   show this help message and exit
    --port PORT  Marker Server port.
    --host HOST  Marker Server host ip.
    --name NAME  Marker Server name.
    --ping       test connection to Markerserver
    --kill       send a poison pill to the Markerserver

"""

import sys
import time
from reiz._marker.mitm import Server
from reiz._marker.client import available, kill
import argparse


def main():

    parser = argparse.ArgumentParser(description="Reiz Marker Server")
    parser.add_argument("--port", dest="port", type=int,
                        help="Marker Server port.", default=7654)
    parser.add_argument("--host", dest="host", type=str,
                        help="Marker Server host ip.", default="127.0.0.1")
    parser.add_argument("--name", dest="name",
                        help="Marker Server name.", default='reiz-marker')
    parser.add_argument("--ping", action="store_true",
                        help="test connection to Markerserver")
    parser.add_argument("--kill", action="store_true",
                        help="send a poison pill to the Markerserver")

    args = parser.parse_args()
    if args.kill:
        if available(host=args.host, port=args.port):
            kill(host=args.host, port=args.port)
        return
    if args.ping:
        response = available(host=args.host, port=args.port)
        if response:
            print(f"Markerserver is available at {args.host}:{args.port}")
            sys.exit(0)
        else:
            sys.exit(1)

    server = Server(port=args.port, name=args.name)
    try:
        server.start()
        while not server.is_running.is_set():
            pass
        if not server.singleton.is_set():
            raise ConnectionAbortedError()
        time.sleep(1)
        print("Server initialized")
        while server.is_running.is_set():
            try:
                pass
            except KeyboardInterrupt:
                break

        server.stop()
        print("\nStreaming stopped.\n")
        server.join()
        print("\nStreaming joined.\n")
        sys.exit(0)

    except ConnectionAbortedError as e:  # erver already connected
        print(e)
        sys.exit(0)


if __name__ == '__main__':

    main()



"""
Create complex shapes and murals
................................
"""
from reiz._visual.colors import get_color, ColorType
import pyglet
from pathlib import Path
from reiz._visual._primitives import Polygon as _Polygon
from reiz._visual._primitives import Circle as _Circle
from reiz._visual._primitives import Line as _Line
from typing import Tuple, NewType, List
# : Tuple[float, float], x-y coordinates scaled from -1 to 1 with 0 indicating the center of the screen
XY = NewType("XY", Tuple[float, float])

# %%


class Visual():
    "The base class for all visual cues to inherit from"

    def adapt(self, window):
        pass

    def draw(self, canvas=None):
        if canvas is not None and not canvas.window.has_exit:
            self.adapt(canvas)
        try:
            for v in self.visual:
                v.draw()
        except TypeError:
            self.visual.draw()

    def set_color(self, color: str):
        self.color = get_color(color)

    def __iter__(self):
        yield self

    def __next__(self):
        yield self
        raise StopIteration

# Complex parametric visualisations
# ------------------------------------------------------------------------------


class Background(Visual):
    """Fills the whole screen with a single color

    args
    ----
    color:str
        the color to fill the screen. see also :data:`reiz.visual.colors.COLORS`

    """

    def __init__(self, color='white'):
        self.color = [int(c*255) for c in get_color(color)]

    def adapt(self, window):
        img = pyglet.image.SolidColorImagePattern(color=self.color)
        img = img.create_image(window.width, window.height)
        self.visual = pyglet.sprite.Sprite(img=img, x=0, y=0,
                                           usage='static')

# ------------------------------------------------------------------------------


class Mural(Visual):
    """A text on the screen

    args
    ----

    text:str
        the text to show
    font:str
        the font-type (needs to be installed on the system)
    fontsize: float
        the normalized size of the letters
    position: XY
        where on the screen the text should be located. Coordinates mark the center of the text
    color: ColorType
        the desired color
    """

    def __init__(self, text: str = 'Hello World', font='Times New Roman',
                 fontsize: float = 1, position: Tuple[float, float] = (0, 0), color=(1, 1, 1),
                 anchor_x='center', anchor_y='center'):
        self.scale = 0.05 * fontsize
        self.text = text
        self.font = font
        self.pos = position
        self.set_color(color)
        self.anchor_x = anchor_x
        self.anchor_y = anchor_y

    def adapt(self, canvas):
        color = tuple(int(c*255) for c in self.color)
        x0 = canvas.width//2
        y0 = canvas.height//2
        x = (x0*self.pos[0]) + x0
        y = (y0*self.pos[1]) + y0
        fontsize = int(canvas.width*self.scale)
        self.visual = pyglet.text.Label(self.text, font_name=self.font,
                                        font_size=fontsize, color=color, x=x, y=y,
                                        anchor_x=self.anchor_x, anchor_y=self.anchor_y)

    def __repr__(self):
        return f"Mural('{self.text}')"

# ------------------------------------------------------------------------------


class Line(Visual):
    """A line from a to b

    args
    ----

    a: XY
        x-y coordinates of the start point
    b: XY
        x-y coordinates of the end point
    color: ColorType
        the desired color
    linewidth:float
    """

    def __init__(self, a: XY = (0, 0), b: XY = (0, 0),
                 color: ColorType = 'white', linewidth: float = 1):
        self.a = a
        self.b = b
        self.color = get_color(color)
        self.linewidth = linewidth

    def adapt(self, window):
        x0 = window.width//2
        y0 = window.height//2
        ax = (x0*self.a[0]) + x0
        ay = (y0*self.a[1]) + y0
        bx = (x0*self.b[0]) + x0
        by = (y0*self.b[1]) + y0

        self.visual = _Line(a=(ax, ay), b=(bx, by), z=0,
                            color=self.color, stroke=self.linewidth)


class Polygon(Visual):
    """An arbitrary shape defined by the coordinates of its nodes

    args
    ----
    positions: List[XY, ]
        a list of x-y-coordinates of each node
    color: ColorType
        the desired color
    """

    def __init__(self, positions: List[XY, ] = [(0, 0), (.1, .2), (.3, .1)], color='white'):
        self.positions = positions
        self.color = get_color(color)

    def adapt(self, window):
        x0 = window.width//2
        y0 = window.height//2

        v = []
        for pos in self.positions:
            x = (x0*pos[0]) + x0
            y = (y0*pos[1]) + y0
            v.append((x, y))
        self.visual = _Polygon(
            v=v, z=0, color=self.color, stroke=0, rotation=0)


class Bar(Visual):
    """A rectangle starting from the bottom centre of the screen

    args
    ----
    height:float
        height in normalized units from 0 to 1 (bottom to top of screen)
    width:float
        width in  normalized units from 0 to 1 (nothing to whole screen)
    color: ColorType
        the desired color
    """

    def __init__(self, height: float = .5, width: float = .25, color='white'):
        self.height = height
        self.width = width
        self.color = get_color(color)

    def adapt(self, window):
        x0 = window.width//2
        y0 = window.height*3/4
        h = self.height
        w = self.width
        positions = [(-w, 0), (-w, h), (w, h), (w, 0)]
        v = []
        for pos in positions:
            x = (x0*pos[0]) + x0
            y = (y0*pos[1]) + 0
            v.append((x, y))
        self.visual = _Polygon(
            v=v, z=0, color=self.color, stroke=0, rotation=0)


class Circle(Visual):
    """A circle, i.e. the outline of a ball

    args
    ----
    zoom: float
        size of the circle
    color: ColorType
        the desired color
    position: XY
        position of the center of the circle
    opacity: float
        how opaque the circle is supposed to be from 0 to 1
    stroke: float
        thickness of the outline
    """

    def __init__(self, zoom: float = 1, color: ColorType = 'red',
                 position: XY = (0, 0), opacity: float = 1, stroke: float = 0):
        self.pos = position
        self.color = get_color(color, opacity)
        self.zoom = zoom
        self.stroke = stroke

    def adapt(self, window):
        x0 = window.width//2
        y0 = window.height//2
        width = int(0.1*min([window.height, window.width]) * self.zoom)

        x = (x0*self.pos[0]) + x0
        y = (y0*self.pos[1]) + y0
        self.visual = _Circle(x=x, y=y, z=0, width=width, color=self.color,
                              stroke=self.stroke)

    def __repr__(self):
        return (f"Circle(zoom={self.zoom}, color={self.color}, " +
                f"position={self.pos})")


class Cylinder(Visual):
    """A rectangular shape which can be rotated around its center

    args
    ----
    position: XY
        position of the center of the circle
    angle:float
        rotation angle around its center
    thickness:float
        width of the cylinder
    length:float
        length of the cylinder
    color: ColorType
        the desired color

    """

    def __init__(self, position=(0, 0), angle: float = 0,
                 thickness: float = .05, length: float = .75,
                 color: ColorType = 'brown'):
        self.color = get_color(color)
        self.length = length
        self.pos = position
        self.angle = angle
        self.thickness = thickness

    def adapt(self, canvas):
        x0 = canvas.width//2
        y0 = canvas.height//2
        x = (x0*self.pos[0]) + x0
        y = (y0*self.pos[1]) + y0
        thickness = int(canvas.height * self.thickness)
        length = int(canvas.width * self.length)

        vertices = ((x-length//2, y-thickness//2),
                    (x-length//2, y+thickness//2),
                    (x+length//2, y+thickness//2),
                    (x+length//2, y-thickness//2))

        self.visual = _Polygon(v=vertices, z=0, color=self.color,
                               stroke=0, rotation=self.angle)


# ------------------------------------------------------------------------------
class Cross(Visual):
    """A fixation cross in the center of the screen

    args
    ----
    zoom: float
        size of the cross
    color: ColorType
        the desired color

    """

    def __init__(self, zoom=1, color='white'):
        self.color = get_color(color)
        self.zoom = zoom

    def adapt(self, window):
        x0 = window.width//2
        y0 = window.height//2
        armlen = int(self.zoom * 100)
        armwid = int(self.zoom * 15)

        y1 = y0-armlen
        y2 = y0-armwid
        y3 = y0+armwid
        y4 = y0+armlen

        x1 = x0-armlen
        x2 = x0-armwid
        x3 = x0+armwid
        x4 = x0+armlen

        v = [(x1, y2), (x1, y3), (x4, y3), (x4, y2)]
        self.ho = _Polygon(v=v, z=0, color=self.color, stroke=0, rotation=0)
        v = [(x2, y1), (x2, y4), (x3, y4), (x3, y1)]
        self.ve = _Polygon(v=v, z=0, color=self.color, stroke=0, rotation=0)

        self.visual = [self.ho, self.ve]

    def __repr__(self):
        return f"Cross(zoom={self.zoom}, color={self.color[:-1]})"


class Trapezoid(Visual):
    """A trapezoid spanned between four nodes

    args
    ----
    xpos: Tuple[float, float, float, float]
        a tuple of x coordinates of the nodes
    ypos: Tuple[float, float]
        a tuple of y coordinates of the nodes
        the first entry is the ypos of the first two xpos
    color: ColorType
        the desired color
    """

    def __init__(self, xpos=(-.33, -.25, .25, .33),
                 ypos=(-.25, .25), color='white'):
        self.xpos = xpos
        self.ypos = ypos
        self.color = get_color(color)

    def adapt(self, window):
        x0 = window.width//2
        y0 = window.height//2

        # generate vertex coordinates
        positions = []
        for x, y in zip(self.xpos[0:2], self.ypos):
            positions.append((x, y))
        for x, y in zip(self.xpos[2:], reversed(self.ypos)):
            positions.append((x, y))

        # create the trapezoid in the first color (default white)
        vertex = []
        for pos in positions:
            x = (x0*pos[0]) + x0
            y = (y0*pos[1]) + y0
            vertex.append((x, y))

        self.visual = _Polygon(
            v=vertex, z=0, color=self.color, stroke=0, rotation=0)

# ------------------------------------------------------------------------------
# %% File-based visualisations


class Image(Visual):
    """Show an image loaded from a file

    args
    ----
    imgpath:str
        path to the imagefile
    position:XY
        x-y-coordinate of the center of the image
    scale:float
        size of the figure relative to the screen size
    """

    def __init__(self, imgpath: str,
                 position: Tuple[float, float] = (0, 0),
                 scale: float = .5):
        self.imgpath = str(Path(imgpath).expanduser().absolute())
        self.img = pyglet.image.load(imgpath)
        self.scale = scale
        self.pos = position

    def adapt(self, window):
        base_scale = min(window.width/self.img.width,
                         window.height/self.img.height)
        scale = self.scale*base_scale
        self.img.scale = scale
        x0 = window.width//2 - (scale * self.img.width//2)
        y0 = window.height//2 - (scale * self.img.height//2)
        x = (x0*self.pos[0]) + x0
        y = (y0*self.pos[1]) + y0

        self.visual = pyglet.sprite.Sprite(img=self.img, x=x, y=y,
                                           usage='static')
        self.visual.scale = scale

    def __repr__(self):
        return f"Image(imgpath='{self.imgpath}')"
# ------------------------------------------------------------------------------



import reiz
import pytest
import time
from pyglet.window import key


def test_canvas_keypress():
    canvas = reiz.Canvas()
    canvas.open()
    assert canvas.start_run == False
    canvas.window.dispatch_event("on_key_press", key.F5, 0)
    canvas.flip()
    assert canvas.start_run == True
    canvas.window.dispatch_event("on_key_press", key.P, 0)
    canvas.flip()
    assert canvas.paused == True
    canvas.window.dispatch_event("on_key_press", key.P, 0)
    canvas.flip()
    assert canvas.paused == False
    canvas.window.dispatch_event("on_key_press", key.ESCAPE, 0)
    canvas.flip()
    assert canvas.available == False


def test_canvas_available(capsys):
    canvas = reiz.Canvas(size=(5, 5))
    assert canvas.available == False
    canvas.open()
    assert canvas.available == True
    canvas.close()
    assert canvas.available == False
    canvas.open()
    assert canvas.available == True
    canvas.close()
    canvas.flip()
    out, err = capsys.readouterr()
    assert "Window was closed" in out
    canvas.open()
    del canvas.window
    canvas.flip()
    out, err = capsys.readouterr()
    assert "Window was closed" in out


def test_canvas_properties(capsys):
    w, h = (1, 1)
    canvas = reiz.Canvas(size=(w, h))
    canvas.open()
    canvas.clear()
    canvas.get_diag()
    canvas.estimate_fps()
    canvas.set_fullscreen()
    assert h < canvas.height
    assert w < canvas.width
    canvas.set_windowed()
    assert h == canvas.height
    assert w == canvas.width
    canvas.close()
    canvas.show(reiz.visual.library.fixation)
    out, err = capsys.readouterr()
    assert "Window was closed" in out
    canvas.open()


def test_clock_sleep():
    limit = 1.5 * 10 ** (-4)  # exact within a millisecond
    desired = 0.1
    actual = reiz.clock.sleep(desired)
    deviance = abs(desired - actual)
    assert deviance < limit


def test_clock_sleep_debiased():
    import random
    import time

    limit = 1.5 * 10 ** (-4)  # exact within a millisecond
    desired = 0.1
    reiz.clock.reset()
    for i in range(1, 11):
        time.sleep(random.random() * 0.05)
        reiz.clock.sleep_debiased(desired)
    actual = reiz.clock.now()
    deviance = abs(desired * i - actual)
    assert deviance < limit



from pytest import fixture
import reiz
import logging

logging.basicConfig(level=1)


@fixture
def rmarker(capsys):
    # set up
    assert reiz.marker.available(verbose=True) is False
    out, err = capsys.readouterr()
    assert "is not available" in out
    assert "Connection refused" in out
    reiz.clock.reset()
    reiz.marker.start()
    assert reiz.clock.now() < 10  # create within 10 secs
    # interrupt
    yield True
    # tear down
    reiz.marker.start()
    reiz.clock.sleep(0.5)
    out, err = capsys.readouterr()
    assert "A marker-server is already running" in out

    reiz.clock.reset()
    reiz.marker.stop()
    assert reiz.clock.now() < 10  # kill within 10 secs

    reiz.marker.stop()
    reiz.clock.sleep(0.5)
    out, err = capsys.readouterr()
    assert "No marker-server is currently running" in out


def test_sanitization(rmarker, capsys):
    assert rmarker
    out, err = capsys.readouterr()
    reiz.marker.push("ä ö ü")
    out, err = capsys.readouterr()
    assert "Sending ae_oe_ue " in out
    out, err = capsys.readouterr()
    reiz.marker.push("ping")
    out, err = capsys.readouterr()
    assert 'Sending {"msg": "ping"}' in out
    out, err = capsys.readouterr()
    reiz.marker.push("poison-pill")
    out, err = capsys.readouterr()
    assert 'Sending {"msg": "poison-pill"}' in out

    c = reiz.Canvas((10, 10))
    reiz.audio.library.beep.volume = 0
    cue = reiz.Cue(
        canvas=c,
        audiostim=reiz.audio.library.beep,
        visualstim=reiz.visual.library.go,
        markerstr="test",
    )

    out, err = capsys.readouterr()
    cue.show()
    out, err = capsys.readouterr()
    assert "Sending test at" in out



"""Client interface to the MarkerServer to send markers"""
import pylsl
import socket
import json
from logging import getLogger
from sys import platform

log = getLogger()


def sanitize_string(marker: str) -> str:
    """sanitize a string

    removes whitespace, transforms to lower case and replaces umlaute and
    spaces with safer symbols

    args
    ----
    marker:str
        the un-sanitized string

    returns
    -------
    sanitized:str
        the sanitized string
    """
    translation = str.maketrans({"ä": "ae", "ö": "oe", "ü": "ue", " ": "_"})
    marker = marker.lower().strip().translate(translation)

    if marker.lower() == "ping":
        marker = json.dumps({"msg": marker})
        log.critical(
            f"'ping' is a reserved key-word for pinging the marker-server. If you want to ping, use available. Forwarded {marker} instead"
        )
    if marker.lower() == "poison-pill":
        marker = {"msg": marker}
        marker = json.dumps(marker)
        log.critical(
            f"'poison-pill' is a reserved key-word for shutting down the marker-server. If you want to shut-down, use kill. Forwarded {marker} instead"
        )
    return marker


def push(marker: str = "", tstamp: float = None, sanitize=True, port: int = 7654):
    """push a marker to the MarkerServer for redistribution as LSL

    args
    ----

    marker: str
        an ascii-encodable string describing an event. We recommend
        to use valid json-strings :func:`~.push_json`
    tstamp: float
        the timestamp of the event. We recommend to use timestamps received
        from pylsl.local_clock
    sanitize: bool
        whether the string is to be sanitized, see :func:`~.sanitize_string`
    port: int
        the port of the MarkerServer

    """
    if tstamp is None:
        tstamp = pylsl.local_clock()
    if sanitize:
        marker = sanitize_string(marker)

    c = _Client(port=port)
    c.push(marker, tstamp)


def push_json(marker: dict = {"key": "value"}, tstamp: float = None):
    """encode a dictionary as json and push it to the MarkerServer

    args
    ----
    marker:
        a json-encodable dictionary describing an event or current settings
    tstamp:
        the timestamp of the event. We recommend to use timestamps received
        from pylsl.local_clock

    .. caution:
        the resulting msg will not be sanitized :func:`~.sanitize_string`
    """
    push(json.dumps(marker), tstamp, sanitize=False)


def available(port: int = 7654, host: str = "127.0.0.1", verbose=True) -> bool:
    """test whether a markerserver is already available at port

    args
    ----

    host: str
        the ip of the markerserver (defaults to localhost)

    port: int
        the port number of the markerserver (defaults to 7654)

    returns
    -------

    status: bool
        True if available, False if not
    """
    c = _Client(host=host, port=port)
    try:
        c.push("ping", pylsl.local_clock())
        return True
    except ConnectionRefusedError as e:
        if verbose:
            print(e)
            print(f"Markerserver at {host}:{port} is not available")
        return False


def kill(host: str = "127.0.0.1", port: int = 7654):
    c = _Client(port=port, host=host)
    c.push("poison-pill", pylsl.local_clock())


class _Client:
    "Basic Client communicating with the MarkerServer"

    def __init__(self, host="127.0.0.1", port: int = 7654, verbose=True):
        self.host = host
        self.port = port
        self.verbose = verbose

    def push(self, marker: str = "", tstamp: float = None):
        "connects, sends a message, and close the connection"
        self.connect()
        self.write(marker, tstamp)
        self.close()

    def connect(self):
        "connect wth the remote server"
        self.interface = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.interface.connect((self.host, self.port))
        self.interface.settimeout(1)

    def write(self, marker, tstamp):
        "encode message into ascii and send all bytes"
        msg = json.dumps((marker, tstamp)).encode("ascii")
        if self.verbose:
            print(f"Sending {marker} at {tstamp}")
        self.interface.sendall(msg)

    def close(self):
        "closes the connection"
        self.interface.shutdown(1)
        self.interface.close()


if "darwin" in platform:  # pragma no cover

    def fake_push(
        marker: str = "", tstamp: float = None, sanitize=True, port: int = 7654
    ):

        if tstamp is None:
            tstamp = pylsl.local_clock()
        if sanitize:
            marker = sanitize_string(marker)
        print(f'MacOs: fake_push "{marker}" at {tstamp}')

    fake_push.__doc__ = push.__doc__
    push = fake_push




"""Cues bind stimuli and a marker for synchronous presentation
"""
import reiz.marker as marker
from reiz.time import Clock as _Clock
from types import SimpleNamespace

class Cue():
    """bind stimuli and a marker for synchronous presentation


    This class binds auditory and visual stimuli together with a canvas
    for presentation and sends markers.


    args
    ----

    canvas: :class:`reiz.Canvas`
        a canvas for presentation of the visual stimuli. Create with
    audiostim: :class:`reiz._audio.primitives.Sound`
        a single auditory stimulus to be presented once at the start
        of :meth:`~.show`.
    visualstim: :class:`reiz._visual.complex.Visual`
        a list or a single visual stimulus. Will be presented on the canvas during the whole duration when :meth:`~.show`. was called. If you use more than one visual stimulus, they will be overlayed and are plotted from left to right. That means the first is at the bottom layer, the last at the top.
    markerstr: str
        a string encapsulating the meaning of the cue. This string will be forwarded to the marker-server, and published with LSL. By default,
        strings are sanitized for easier parsing, so do not use case-sensitive
        information, and try to limit yourself to ascii.

    """

    def __init__(self, canvas=None, audiostim=None,
                 visualstim=None, markerstr=None):
        self.canvas = canvas
        self.audio = audiostim
        self.visual = visualstim
        self.marker = markerstr

    def show(self, duration: float = 0, canvas=None, safetime=.2):
        """present all stimuli stored in the cue

        args
        ----
        duration: float
            how many seconds you want to present the visual stimuli on the canvas. Defaults to zero.
            Using zero will result in the function returning immediatly, while
            keeping the visual stimulus maintained on the canvas. In that regard, it resembles presentation forever, (or until another cue is presented). This can confuse Windows 10 if the canvas is moved or resized, as the OS thinks the canvas is `unresponsive`. Any other positive values cause the function to not return and block for the whole duration.
        canvas: :class:`reiz.Canvas`
            In case you decide to present the Cue on a different canvas instead to the one assigned during instantiation. Will be ignored otherwise.
        safetime: float
            continuous presentation causes the canvas to be updated at the flip rate of your screen and grapics cards (usually aroung 60Hz). This hardware limitation means that the duration of presentation
            will be quantizised, i.e. the duration of presentation will be a multiple of 16.6ms for a 60Hz screen. To achieve more accurate duration, the safetime parameters sets how long to the end of the duration we will no longer flip the screen. By default, we won't flip the screen during the last 200ms. This can cause the OS to consider the screen `unresponsive` (see duration above). Please note that this does not improve the timing of the actual drawing on the screen, it just allows the :meth:`~.show` to return at a more accurate time.

        """

        if canvas is not None:
            self.canvas = canvas
        if self.audio is not None:
            self.audio.play()
        if self.marker is not None:
            marker.push(self.marker)

        if duration is not None and self.visual is not None:
            if duration == 0:  # show "forever"
                self.canvas.show(self.visual)
                return 0
            else:  # show for duration
                clk = _Clock()
                dt = clk.now()
                clk.tick()
                self.canvas.show(self.visual)
                while dt <= duration and self.canvas.available:
                    # flipping quantizises the sleep duration, therefore we
                    # don't flip anymore if that would become relevant
                    # we only repeat presentation of the visual stimulus
                    # as audio and markers would be senseless and overlays
                    if abs(duration-dt) > safetime:
                        self.canvas.show(self.visual)
                        clk.sleep_debiased(.1)
                    dt = clk.now()
                return dt


def collect(**lib) -> SimpleNamespace:
    lib = SimpleNamespace(**lib)
    return lib



import reiz
from pytest import fixture


@fixture
def stimuli(canvas):
    cue = reiz.Cue(
        canvas=canvas,
        audiostim=reiz.audio.library.beep,
        visualstim=reiz.visual.library.go,
    )
    N = reiz.Cue(canvas=canvas, audiostim=None, visualstim=None)
    return cue, N


def test_cue(canvas, stimuli):
    cue, N = stimuli
    reiz.audio.library.beep.volume = 0
    cue.show(0.5)
    N.show(canvas=canvas)
    N.show()
    cue.show(None)


def test_collection(stimuli):
    cue, N = stimuli
    lib = reiz.cue.collect(**{"N": N, "C": cue})
    assert "N" in lib.__dict__
    assert "C" in lib.__dict__
    assert lib.N is N
